/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "fe/platform.h"

namespace fe
{

FE_DL_PUBLIC MutexInitFunction* Mutex::ms_fnInit=NULL;
FE_DL_PUBLIC MutexLockFunction* Mutex::ms_fnLock=NULL;
FE_DL_PUBLIC MutexFinishFunction* Mutex::ms_fnFinish=NULL;

FE_DL_PUBLIC MutexGuardInitFunction* Mutex::ms_fnGuardInit=NULL;
FE_DL_PUBLIC MutexGuardLockFunction* Mutex::ms_fnGuardLock=NULL;
FE_DL_PUBLIC MutexGuardFinishFunction* Mutex::ms_fnGuardFinish=NULL;

FE_DL_PUBLIC MutexConditionInitFunction* Mutex::ms_fnConditionInit=NULL;
FE_DL_PUBLIC MutexConditionWaitFunction* Mutex::ms_fnConditionWait=NULL;
FE_DL_PUBLIC MutexConditionNotifyFunction* Mutex::ms_fnConditionNotify=NULL;
FE_DL_PUBLIC MutexConditionFinishFunction* Mutex::ms_fnConditionFinish=NULL;

FE_DL_PUBLIC BWORD Mutex::ms_checked=FALSE;

Mutex::Mutex(void):
	m_recursive(FALSE),
	m_initializing(FALSE),
	m_pRawMutex(NULL)
{
	//* defer init
}

Mutex::~Mutex(void)
{
	if(m_pRawMutex)
	{
		if(ms_fnFinish)
		{
			ms_fnFinish(m_recursive,m_pRawMutex);
		}
		m_pRawMutex=NULL;
	}
}

void* Mutex::raw(void)
{
	if(!m_pRawMutex && !m_initializing)
	{
		confirm();

		if(ms_fnInit)
		{
			m_initializing=TRUE;
			m_pRawMutex=ms_fnInit(m_recursive);
			m_initializing=FALSE;
		}
	}

	return m_pRawMutex;
}

BWORD Mutex::lock(void)
{
	if(!ms_fnLock || m_initializing)
	{
		return FALSE;
	}

	return ms_fnLock(m_recursive,raw(),false,false,false);
}

BWORD Mutex::unlock(void)
{
	if(!ms_fnLock || m_initializing)
	{
		return FALSE;
	}

	return ms_fnLock(m_recursive,raw(),true,false,false);
}

BWORD Mutex::tryLock(void)
{
	if(!ms_fnLock || m_initializing)
	{
		return FALSE;
	}

	return ms_fnLock(m_recursive,raw(),false,true,false);
}

BWORD Mutex::lockReadOnly(void)
{
	if(!ms_fnLock || m_initializing)
	{
		return FALSE;
	}

	return ms_fnLock(m_recursive,raw(),false,false,true);
}

BWORD Mutex::unlockReadOnly(void)
{
	if(!ms_fnLock || m_initializing)
	{
		return FALSE;
	}

	return ms_fnLock(m_recursive,raw(),true,false,true);
}

BWORD Mutex::tryLockReadOnly(void)
{
	if(!ms_fnLock || m_initializing)
	{
		return FALSE;
	}

	return ms_fnLock(m_recursive,raw(),false,true,true);
}

Mutex::Guard::Guard(Mutex& a_rMutex,BWORD a_readOnly):
	m_pMutex(&a_rMutex),
	m_pRawGuard(NULL),
	m_readOnly(a_readOnly)
{
	confirm();

	if(Mutex::ms_fnGuardInit)
	{
		void* pRawMutex=m_pMutex->raw();
		if(pRawMutex)
		{
			m_pRawGuard=Mutex::ms_fnGuardInit(
					m_pMutex->m_recursive,pRawMutex,m_readOnly);
			FEASSERT(m_pRawGuard);
		}
	}
}

Mutex::Guard::~Guard(void)
{
	if(Mutex::ms_fnGuardFinish && m_pRawGuard)
	{
		Mutex::ms_fnGuardFinish(m_pMutex->m_recursive,m_pRawGuard,m_readOnly);
	}
}

BWORD Mutex::Guard::lock(void)
{
	if(Mutex::ms_fnGuardLock && m_pRawGuard)
	{
		return Mutex::ms_fnGuardLock(
				m_pMutex->m_recursive,m_pRawGuard,FALSE,FALSE,m_readOnly);
	}
	return FALSE;
}

BWORD Mutex::Guard::unlock(void)
{
	if(Mutex::ms_fnGuardLock && m_pRawGuard)
	{
		return Mutex::ms_fnGuardLock(
				m_pMutex->m_recursive,m_pRawGuard,TRUE,FALSE,m_readOnly);
	}
	return FALSE;
}

BWORD Mutex::Guard::tryLock(void)
{
	if(Mutex::ms_fnGuardLock && m_pRawGuard)
	{
		return Mutex::ms_fnGuardLock(
				m_pMutex->m_recursive,m_pRawGuard,FALSE,TRUE,m_readOnly);
	}
	return FALSE;
}

Mutex::Condition::Condition(void):
	m_pRawCondition(NULL)
{
	if(Mutex::ms_fnConditionInit)
	{
		m_pRawCondition=Mutex::ms_fnConditionInit();
	}
}

Mutex::Condition::~Condition(void)
{
	if(Mutex::ms_fnConditionFinish)
	{
		Mutex::ms_fnConditionFinish(m_pRawCondition);
	}
}

BWORD Mutex::Condition::wait(Guard& a_rGuard)
{
	if(Mutex::ms_fnConditionWait)
	{
		void* pRawGuard=a_rGuard.m_pRawGuard;
		if(pRawGuard)
		{
			return Mutex::ms_fnConditionWait(m_pRawCondition,
					a_rGuard.m_pMutex->m_recursive,pRawGuard,
					a_rGuard.m_readOnly);
		}
	}
	return FALSE;
}

BWORD Mutex::Condition::notifyOne(void)
{
	if(Mutex::ms_fnConditionNotify)
	{
		return Mutex::ms_fnConditionNotify(m_pRawCondition,FALSE);
	}
	return FALSE;
}

BWORD Mutex::Condition::notifyAll(void)
{
	if(Mutex::ms_fnConditionNotify)
	{
		return Mutex::ms_fnConditionNotify(m_pRawCondition,TRUE);
	}
	return FALSE;
}

//* static
BWORD Mutex::confirm(void)
{
	if(ms_fnInit)
	{
		return TRUE;
	}
	if(ms_checked)
	{
		return FALSE;
	}
	ms_checked=TRUE;

	printf("Free Electron\n");

	DL_Loader *pLoader=new DL_Loader();
	FEASSERT(pLoader);

	const I32 nameCount=4;
	const char* names[nameCount]=
			{ "fexBoostThread1_37", "fexBoostThread1_46",
			"fexBoostThread1_55", "fexBoostThread" };

	BWORD success=FALSE;
	I32 nameIndex=0;
	const char *name=NULL;

	while(!success && nameIndex<nameCount)
	{
		name=names[nameIndex];
		const BWORD adaptname=TRUE;
		success=pLoader->openDL(name,adaptname);
		nameIndex++;
	}

	if(!success)
	{
		fprintf(stderr,"  failed to find mutex support\n");
		delete pLoader;
		return FALSE;
	}

	bool (*mutex_init)(void);
	mutex_init=(bool (*)(void))pLoader->findDLFunction("mutex_init");

	if(!mutex_init)
	{
		fprintf(stderr,"  mutex_init not found\n");
		delete pLoader;
		return FALSE;
	}

	printf("  using %s (for mutexes)\n",name);

	return mutex_init();

	//* TODO retain, close, and delete pLoader later
}

} /* namespace */

